/*
  
  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/
  
  ===================================================
  
  Name:       CCM (bookKeeping.C)
  
  Copyright:  Damian J. J. Farnell
  
  Date:       Wed Sep 29 1999
  
  Purpose:    Determine and solve CCM LSUBm equations.

  Rewritten:  Joerg Schulenburg 2005-2009
  
  ===================================================
  
  ---------------------------------------------------

  MODULE NAME: BOOKKEEPING.C 

  PURPOSE:     This module does all of the `bookkeeping' 
               for all of the other modules. Hence, all
               routines that are multiply used throughout
               the code are contained here. 

  ---------------------------------------------------

*/


// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

#include "nr.h"
#include "nrutil.h"
#include "Constants.h"
#include "Externals.h"
#include "FunctionPrototypes.h"
// #define DEBUG 1

// return the number of realtime seconds since start
int seconds( void ) {
  static time_t start_time=0;
  time_t t1;
  time(&t1); if (start_time==0) start_time=t1;
  return (t1-start_time);
}

void printOutConfig(int (*conf)[5], int order) {
  int i;

  // we make the output as compact as possible (x [y [z [uc]]])
  // output as list
  if (0<Params.dim) printf(" x");
  if (1<Params.dim) printf(" y");
  if (2<Params.dim) printf(" z");
  if (UnitCell.number > 1) printf(" uc");
  printf(" [%2d]:", order);
  for (i=1; i<=order; i++) {
    if (0<Params.dim) printf(" %d", conf[i][1]);
    if (1<Params.dim) printf(" %d", conf[i][2]);
    if (2<Params.dim) printf(" %d", conf[i][3]);
    if (UnitCell.number > 1) printf(" %d", conf[i][4]);
    if (i<order) printf(",");
  }
  printf("\n");
}

// JS 2009-07: output patterns within unitcell too
void printOutPattern(int (*conf)[5], int order) {
  int i,j1,j2,j3,max[3];
  int flips=0;
  max[0]=max[1]=max[2]=0; /* covering cube */

  for (i=1; i<=order; i++) {
    if (max[0]<conf[i][1]) max[0]=conf[i][1];
    if (max[1]<conf[i][2]) max[1]=conf[i][2];
    if (max[2]<conf[i][3]) max[2]=conf[i][3];
  }
  // output as pattern
  if (max[2]<1) { /* flat animal? */
    for (j1=0;j1<max[1]+1;j1++) {
      printf(" ");
      for (j2=0;j2<max[0]+1;j2++) {
        for (j3=1;j3<=UnitCell.number;j3++) {
          if (UnitCell.number>1 && j3==1) printf("("); // UC-frame
          flips=0;
          for (i=1; i<=order; i++) {  // slow but not speed relevant
            if (conf[i][4]==j3 && conf[i][2]==j1 && conf[i][1]==j2) flips++;
          }
          printf("%c",((flips)?'0'+flips:'.'));
          if (UnitCell.number>1 && j3==UnitCell.number) printf(")");
        }
      }
      printf("\n");
    }
  }
}

void printOutDescriptor(int *descriptor) {
  int i, j;
  printf("descriptor[0..%d]:  ",DESCR_SIZE-1);
  for(j=DESCR_SIZE-1;j>=0;j--) if (descriptor[j]) break;
  if (j<DESCR_SIZE) j++;
  if (j<DESCR_SIZE) j++;
  for(i=0;i<j;i++) 
    printf("%d ",descriptor[i]);
  if (j<DESCR_SIZE) printf("...");
  printf("\n");
}


// storage place for (Params.s2)^(j-1) to speedup code
// sum of all divisors must fit into integer (for spin-bit pattern)
// int *divisors=NULL;  // multiplicators (2s)^(i-1) for each site i
int sites_per_int=0; // old: block_size, number of sites fitting to an integer
int bits_per_site=0; // Nov09

void initDescr(int s2) {  // Oct05: s2=Params.s2 Nov09: old: initDivisors
  int tempS, j, prod;

  tempS = s2+1;  // 2s+1 states fitting in the code
  // count bits needed to store all spin states
  for(j=0,prod=tempS; prod; prod>>=1, j++); bits_per_site=j;

  if (tempS<2) {printf("ERROR initDescr: 2s<1 (endless loop)\n"); exit(1);}

  // calculate the maximum number of sites fitting to an integer
  for(j=0,prod=1; ; prod*=tempS, j++) {   // tempS ^ (j)
    if ((prod * tempS) * 1.0 != (prod * 1.0) * tempS) break;
  }
  for(j=0,prod=1<<bits_per_site; prod>0 ; prod<<=bits_per_site, j++);
  sites_per_int=j;    // BLOCK_LENGTH = j;
  // ToDo: replace divisors /% by bits_per_site >>,& for MIPS and easier coding?
  // 32 s=1/2 sites per 32bit (2,4,8, ... or simpler 32*1bit size+0%)
  // 19 s=1   sites per 32bit (3,9,27,... or simpler 16*2bit +16%)
  // 15 s=3/2 sites per 32bit (4,16,64... or simpler 10*3bit +33%)
  // 13 s=4/2 sites per 32bit (5,25,125.. or simpler 10*3bit +23%)
  // 11 s=5/2 sites per 32bit (6,36,...   or simpler 10*3bit +9%)

  if (myid==0)
    printf("\ninitDescr:\n"
           " maximum sites fitting to int is sites_per_int= %d\n"
           " bits needed for %d spin states is bits_per_site= %d\n\n",
            sites_per_int, s2+1, bits_per_site);

#if 0 // JS0911
  // DESCR_SIZE*sites_per_int >= MAX_BT*uc // ToDo: check max_spin_list
  divisors = (int *) malloc(sizeof(int)*(sites_per_int+1));
  if (!divisors) { printf("Error initDescr: malloc failed\n"); exit(1); }
  divisors[0] = 1;
  for(j=0, prod=1; j<=sites_per_int; prod*=tempS, j++) {
    divisors[j] = prod;  // = (2s) ^ (j)
    // printf("DBG: divisors[%d]= %d\n", j, divisors[j]); 
  }
#endif
}

// Apr08 JS
extern int subsetClusters(int (*config)[5], int length); // 1 if ok, 0 if not
extern int subsetExcitedClusters(int (*config)[5], int length); // 1 if ok, 0 if not
extern int flag_subset;

// in: order of ccm (maximum spinflips)
//     configuration[1...order?][5]
// out: descriptor[DESCR_SIZE] of smallest symmetric config
// cost most CPU time (Joerg Oct2004)
// called by findFundamentalSpinConfigurations() in define.C
//       and termXX() in obtain.C
// return 1, if symmetry operations can not lower the descriptor (Nov05) ???
// return -1 on error
//    0 = ok
// ToDo: melt with findExcitedDescriptorOfCluster!
// Joerg Oct09:  remove 999999 flag, buggy for 2^(n=20) > 999999
int findDescriptorOfCluster(int order,
			    int (*configuration)[5],
			    int *descriptor){

  /* This routine gets a set of crystallographic lattice sites
     which define a cluster and returns a set of unique integer 
     numbers in `descriptor' which fully describe it.

     Note that this is its FUNDAMENTAL description, i.e., 
     of the how ever many possible symmetry equivalent 
     cluster there are of it, this `descriptor' has 
     the lowest numerical value.

     PARTS ARE:

     1) determine symmetry related equivalents and translate
     ...into the x>=0,y>=0,z>=0 part.

     2) determine the `spin' form of these equivalents

     3) turn this `spin' form into a unique set of integers

     4) find set of integers with lowest `numerical' value
     ...out of all the symmetry related clusters and return
     ...this in `descriptor'.

  */

  int i,j,k;
  int minR[3];  // x y z
  int OK, 
      is_minConfig, // -1: error (no valid cfg), 1: was min cfg? 
      err=0;  // have valid descriptor
  static int n_out=0;  // ToDo: reset if &(*configuration) changes
  // int **tempConfiguration;  // obsolete
  static int tempDescriptor[DESCR_SIZE],
             t999Descriptor[DESCR_SIZE]; // detect old errors
  static int err999n=0,err999m=1; // detect v091013pre-999999 errors

  // tempConfiguration=matrix(0,MAX_LSUBM+1,0,5);  // to slow
  static int tempConfiguration[MAX_LSUBM+1][5];

  is_minConfig=1;  // no sym operations needed?
  // initialize descriptor? for min/max search
  for (k=0;k<DESCR_SIZE;k++) descriptor[k]=0;  
  for (k=0;k<DESCR_SIZE;k++) t999Descriptor[k]=0;
  t999Descriptor[0]=t999Descriptor[DESCR_SIZE-1]=999999;

#if 1 /* now in findClusterIndex too */
  if (flag_subset && 
      !subsetClusters(configuration,order)) // JS Apr08 SpeedUP????
    {  return -1;}
#endif
 
  if (n_out<=100) n_out++;

  if (Control.output >4 && n_out<100)
      printf("findDescriptorOfCluster:\n");
  
  // ToDo: check sub-m and sub-n here! JS0903 (term27,30,...)

  // we have at least 1 symmetry (Identity-Operator) here
  for(i=1;i<=LatticeSymmetries.number;i++) {

    // find symmetry related clusters

    symmetryGroup(i,order,configuration,tempConfiguration);

    // translate into correct quadrant, x>=0,y>=0,z>=0
    
    for (k=0; k<Dimension && k<Params.dim; k++) {
      minR[k]=tempConfiguration[1][k+1];
      for (j=2; j<=order; j++)
        if (minR[k] > tempConfiguration[j][k+1]) 
            minR[k] = tempConfiguration[j][k+1];
      for (j=1; j<=order; j++)
        tempConfiguration[j][k+1] -= minR[k];
      if (is_minConfig && i==1 && minR[k]) is_minConfig=0;
    }

    // now find the descriptor (costs most CPU, Joerg Oct04)
    // bitpattern (flips per site.uc) splittet to an integer array
    
    err= findConfigurationToDescriptor(tempDescriptor,
                                       tempConfiguration, order);
    // -1 on error, 0 = OK
    if (err<0 && i==1) break; // JS091015
    if (err<0) { // JS0910 Oct09  we have a invalid cluster
#if DEBUG      // reduce code for L1 cache
       static int nerr2=0;
       if (Control.output>7 && nerr2++<100) // term 27 + term 30
         printf("Error findDescriptorOfCluster: invalid descriptor,"
              " sym=%d\n", i);
#endif
       continue;
    }

    // check if tempDesc is smaller than Desc
    // could be bad if valid descr[k] > 999999 (v091013--), fixed by $err
    // check for buggy version result (compare with fixed version later)
    for (OK=0,k=(DESCR_SIZE-1); k>=0; k--) {
      if (tempDescriptor[k] > t999Descriptor[k]) { OK=-1; break; }
      if (tempDescriptor[k] < t999Descriptor[k]) { OK= 1; break; }
    }
    if ((OK==1 || i==1) && err!=-1) { // tempDescr<descr: input config will be replaced
      for (k=0;k<DESCR_SIZE;k++) t999Descriptor[k]=tempDescriptor[k];
    }

    for (OK=0,k=(DESCR_SIZE-1); k>=0; k--) {
#if DEBUG                                
      if (Control.output>10 && (descriptor[k] || tempDescriptor[k]))
        printf("DBG descr_min: sym k %d %d %6d %6d\n", i, k, 
               descriptor[k], tempDescriptor[k]);
#endif
      
      if (tempDescriptor[k] > descriptor[k]) { OK=-1; break; }
      if (tempDescriptor[k] < descriptor[k]) { OK= 1; break; }
    }

    if ((OK==1 || i==1) && err!=-1) { // tempDescr<descr:  input config will be replaced
      if (i!=1) is_minConfig = 0;  // have at least one valid descriptor found 
      for (k=0;k<DESCR_SIZE;k++) descriptor[k]=tempDescriptor[k];
#if DEBUG                                
      if (Control.output>10)
        printf("DBG descr_min: set new min sym %d dscr[0]=%d\n", i, 
               descriptor[0]);
#endif
    }
    if (is_minConfig && OK && i==1 && err!=-1) is_minConfig=1;
    if (i==1 && err==-1) is_minConfig=-1;

  }

  // compare fixed with buggy version
  for (k=0;k<DESCR_SIZE && t999Descriptor[k]==descriptor[k]; k++);
  if (k<DESCR_SIZE
   && !(t999Descriptor[0]==999999 && descriptor[0]==0)) // DEBUG
   {
     err999n++;
     if (err999n == err999m && err999m) {
       err999m<<=1;
       printf("WARNING: old 999999-bug v091013-- (now ok) sym=%d num=%d\n", i, err999n);
       printOutDescriptor(     descriptor );
       printOutDescriptor( t999Descriptor );
       printOutConfig(    configuration, order);
       printOutConfig(tempConfiguration, order);
     }
   }

#if DEBUG
  if (is_minConfig<0) {
    static int nerr2=0;
    // can happen for generated configs with maxdist m > Params.m (JS0903)
    printf("Error: no reasonable descriptor found! Exiting?\n");
    printf("isMinCfg= %d, Minimum Descriptors ", is_minConfig);
    printOutDescriptor( descriptor );
    printOutConfig(configuration, order);
    printOutConfig(tempConfiguration, order);
    // its not necessary a desaster to return here, but we waste compute time
    Control.output=3; // we want to know why this happens and debug
    if (++nerr2==200) exit(1);
  }
#endif
  if (Control.output > 4 && n_out<100) {
    printf(" Minimum Descriptors = ");
    printOutDescriptor( descriptor );
  }
  // free_matrix(tempConfiguration,0,MAX_LSUBM+1,0,5);

  return is_minConfig; // JS Nov05: to count configs with low memory Oct09???
}

// ToDo: melt with findDescriptorOfCluster!
//   only differs in  subset[Excited]Clusters Jun08
int findExcitedDescriptorOfCluster(int order,
				   int (*configuration)[5],
				   int *descriptor){

  int ret, save_flag_subset= flag_subset; // JS0910 bad construct, clean it!
  if (flag_subset && 
      !subsetExcitedClusters(configuration,order)) // JS Apr08 SpeedUP????
    {  return -1; }
  flag_subset = 0;
  ret = findDescriptorOfCluster(order, configuration, descriptor);
  flag_subset = save_flag_subset;
  return ret;
 
}

// cartesian coord x[0..2] to integer config=(x,y,z,uc)
static int findGeneralisedPositionFromCartesians(double x[3], int *config)
{
  int i, j1;
  double distance = 0.0;
  static double s[4], r[3];

  // ToDo: better coding UC shift in symmetry directly
  config[4]=-1;
  for(i=1;i<=UnitCell.number;i++) {

    r[0] = x[0] - UnitCell.x[i];
    r[1] = x[1] - UnitCell.y[i];
    r[2] = x[2] - UnitCell.z[i];

    // Solves the set of n linear equations A x = b (here: tempm.x = s)
    //         B = A^-1   B.A.x = x = B.b
    // only needed for non orthogonal bravais lattices
    // ToDo: rotate xyz outside loop and search i for min dist s[]-UC.xyz[i]
    for (j1=0; j1<Dimension; j1++)
      s[j1+1] = inversBravaisLattice[0][j1] * r[0]
#if Dimension > 1
              + inversBravaisLattice[1][j1] * r[1]
#if Dimension > 2
              + inversBravaisLattice[2][j1] * r[2]
#endif
#endif
    ;
    // check if the values for s[] are integers.
    // round: -0.9 to -1, +0.9 to 1.0  ((int)-2.8 = -2)
    for (j1=0; j1<Dimension; j1++)
      config[j1+1]= (int)(s[j1+1]+((s[j1+1]<0)?-.5:+.5));
//    config[j1+1]= -(1<<30)+(int)(s[j1+1]+.5+(1<<30)); // 2.8*slower @MIPS
    config[4]= i;   // UnitCell site index

    distance= 0;
#if MAX_UNIT_CELL > 1
    for (j1=0; j1<Dimension; j1++)
      distance += fabs( r[j1] - config[1]*BravaisLattice[0][j1]
#if Dimension > 1
                              - config[2]*BravaisLattice[1][j1]
#if Dimension > 2
   	   	              - config[3]*BravaisLattice[2][j1]
#endif
#endif
                    );
    if (distance < 0.001) break;   // if integers, its found
#else
    break;
#endif  // MAX_UNIT_CELL > 1
  }  // i = 1...UnitCell

#if MAX_UNIT_CELL > 1
#if DEBUG
  if (distance >= 0.001)  // if no integer 
    printf("Error findGeneralisedPositionFromCartesians: no UC-site found\n"
           " out: (% 2d % 2d % 2d %d) dist= %g\n"
           " in: xyz= % 2.0f % 2.0f % 2.0f\n", 
           config[1], config[2], config[3], config[4],
           distance, x[0], x[1], x[2]);
#endif
  if (distance >= 0.001) return -1;  // if no integer 
#endif
  return 0;
}


// in:  sym = symmetry_number 1..48
//      order - length of configuration (ToDo: use stop code? config[4]<0?)
//      configuration[1..order][1..4] 
// out: tempConfiguration[1..order][1..4] - symmetric coords 
//  ToDo: bravais -> cartesian -> c_sym -> cartesioan -> bravais
//                 to  bravais -> b_sym -> bravais (integer only!?)
//        1D: speedup=2
void symmetryGroup(int sym, 
		   int order,
		   int (*configuration)[5],
		   int (*tempConfiguration)[5])
{
  int i;

#if defined(SymInt2D) // measure possible speedup 2D square (Oct09 speedup=2)
  int bsym[8][2][2]={{{ 1, 0},{ 0, 1}},
                     {{-1, 0},{ 0, 1}},
                     {{ 1, 0},{ 0,-1}},
                     {{-1, 0},{ 0,-1}},
                     {{ 0, 1},{ 1, 0}},
                     {{ 0,-1},{ 1, 0}},
                     {{ 0, 1},{-1, 0}},
                     {{ 0,-1},{-1, 0}},
                    }; // nosym + reflection
  if (UnitCell.number!=1 || Params.dim!=2) ERR(999);
  for (i=1;i<=order;i++) {
    tempConfiguration[i][1]=bsym[sym-1][0][0]*configuration[i][1]
                           +bsym[sym-1][0][1]*configuration[i][2];
    tempConfiguration[i][2]=bsym[sym-1][1][0]*configuration[i][1]
                           +bsym[sym-1][1][1]*configuration[i][2];
    tempConfiguration[i][3]=                  configuration[i][3];
    tempConfiguration[i][4]=                  configuration[i][4];
  }
#elif defined(SymInt1D) // measure possible speedup 1D
  int bsym[2][1]={{1},{-1}}; // nosym + reflection
  if (UnitCell.number!=1 || Params.dim!=1) ERR(999);
  for (i=1;i<=order;i++) {
    tempConfiguration[i][1]=bsym[sym-1][0]*configuration[i][1];
    tempConfiguration[i][2]=               configuration[i][2];
    tempConfiguration[i][3]=               configuration[i][3];
    tempConfiguration[i][4]=               configuration[i][4];
  }
#else

  double newx[3];
  double x[3];

  // symmorphic point group symmetries entered here only =>
  // triangular lattice not-yet implemented, to be done. (JS0903: ???)

#ifdef DEBUG
  if (sym > 48 || sym < 1) {
    printf("Switch Variable out of range %d...\nExiting!\n\n", sym);
    MPI_Abort(MPI_COMM_WORLD, 12);
  }
#endif

// JS091015 replace B2C*SYM_C*C2B by SYM_B (B=bravais C=cartesian coordinates)
  for(i=1;i<=order;i++) {
    
    // configuration[x,y,z,uc]_Bravais -> (x,y,z)_Cartesian
    findCartesiansFromGeneralisedPosition(&x[0], &x[1], &x[2], configuration[i]);

// to improve performance, we use Dimension
    newx[0]=
#if Dimension > 0
         x[0]*LatticeSymmetries.R[sym][1][1]+
#if Dimension > 1
         x[1]*LatticeSymmetries.R[sym][1][2]+
#if Dimension > 2
         x[2]*LatticeSymmetries.R[sym][1][3]+
#endif
#endif
#endif
           LatticeSymmetries.R[sym][1][4];
    
    newx[1]=
#if Dimension > 1
         x[0]*LatticeSymmetries.R[sym][2][1]+
         x[1]*LatticeSymmetries.R[sym][2][2]+
#if Dimension > 2
         x[2]*LatticeSymmetries.R[sym][2][3]+
#endif
#endif
           LatticeSymmetries.R[sym][2][4];
    
    newx[2]=
#if Dimension > 2
         x[0]*LatticeSymmetries.R[sym][3][1]+
         x[1]*LatticeSymmetries.R[sym][3][2]+
         x[2]*LatticeSymmetries.R[sym][3][3]+
#endif
           LatticeSymmetries.R[sym][3][4];
  
    // ToDo: this consumes most of the time Sep05
    // restore tempConfig=(x,y,z,uc)_bravais from (newx,newy,newz)_cartesian
    //  ToDo: use new[3]
    if (findGeneralisedPositionFromCartesians(newx, tempConfiguration[i])) {
#if DEBUG
       // Fatal Error: not found
       printf("Error symmetryGroup: sym=%d site_of_cfg=%d\n", sym, i);
       printOutConfig(configuration, order);
       printOutPattern(configuration, order);
#endif
       // exit(1); // Oct05 ToDo: probably its just because of inaccuracy
    }
  }
#endif
}

// in:  x y z
// out: configuration[1..4]
// called: by symmetryGroup()  
// return 0 on success, otherwise fatal error
// ToDo: called very often by obtain! takes much cpu time, Joerg Sep05
//   cartesian coordinates to Bravais coordinates + site within unitcell

// in:  configuration[5]
// out: x, y, z
// called by bookKeeping.C and define.C
void findCartesiansFromGeneralisedPosition(double *x, 
					   double *y, 
					   double *z, 
					   int *configuration)
{

  // this routine finds the cartesian position from the 
  // `generalised' lattice position => bravais vector + primitive
  // cell vector

  *x=
#if Dimension > 0
     BravaisLattice[0][0]*configuration[1]+
#if Dimension > 1
     BravaisLattice[1][0]*configuration[2]+
#if Dimension > 2
     BravaisLattice[2][0]*configuration[3]+
#endif
#endif
#endif
             UnitCell.x[configuration[4]];

  *y=
#if Dimension > 1
     BravaisLattice[0][1]*configuration[1]+
     BravaisLattice[1][1]*configuration[2]+
#if Dimension > 2
     BravaisLattice[2][1]*configuration[3]+
#endif
#endif
              UnitCell.y[configuration[4]];

  *z=
#if Dimension > 2
     BravaisLattice[0][2]*configuration[1]+
     BravaisLattice[1][2]*configuration[2]+
     BravaisLattice[2][2]*configuration[3]+
#endif
              UnitCell.z[configuration[4]];

}


// coding the configuration into a descriptor index (splitted to integers)
// out: descriptor[0..DESCR_SIZE-1]=...,999999(=invalid)
// in:  config[1..order][1..4]=0..btLength for [1..3], ??? for [4]
// called by findDescriptorOfCluster() in bookKeeping.C
// not a good name I think (Joerg05)
// JS0910: replace flag 999999 by return_ok_code (0=ok, -1: failed)
// ToDo: rename to get DescrFromConfig() JS0911
int findConfigurationToDescriptor(int *descriptor,int (*config)[5],int order)
{
  int i,j,sum_abs;
  int index;
  // int tempNum;
  int block;

  for(i=0;i<DESCR_SIZE;i++) descriptor[i]=0;  // reset descriptor bits
  
  // the first part is to determine a unique lattice index in the bounding
  // triangle for each spin flip operator in a given cluster. Each spin
  // flip is denoted by it bravais lattice vector + its number in the 
  // unit cell. 
  
  // each cell in the the bounding triangle is a unit cell.

  for(i=1;i<=order;i++) {

    // check if within bounding triangle in 1D, 2D and 3D
    for (sum_abs=0, j=1; j<=Params.dim; j++) sum_abs += abs(config[i][j]);
    if (sum_abs > Params.btLength-1 ) {
      // point lies outside the bounding triangle.
      // happens only for term27+term30 where a site is added 
#ifdef DEBUG
      if (Control.output > 10) {
        printf("Error findConfigurationToDescriptor:"
              " Point outside the bounding triangle (i=%d dist=%d max=%d)\n",
              i, sum_abs, Params.btLength);
        printf("Source config:\n");
        printOutConfig(config,order);
        printOutPattern(config,order);
      }
#endif
      return -1;
    }

#ifdef DEBUG
    if (config[i][1]>MAX_LSUBM
     || config[i][2]>MAX_LSUBM
     || config[i][3]>MAX_LSUBM) {
      printf("Error: config > MAX_LSUBM\n");
      exit(1);
    }
#endif
    // get a unique index for every possible site (see lookUpTable)
    // compression of configuration (save memory)

    // old: index = BTarray[ z ][ y ][ x ];
    index = BTarray[config[i][3] * Params.btLength + config[i][2]]
                                                   + config[i][1];

    // at this point: 0 <= index < max_numbers_cells_in_BT

    // we now have an unique site ``index'' for 1D-3D which completely
    // specifies the position of a given spin in the bounding triangle.
    // we now wish to encode this information in a given ``block'' 
    // with a given number ``tempNum'' in ``descriptor''.

    // note that sites_per_int gives the numbers of ``spins'' the
    // each element of the descriptor describes. Normally set to
    // 20 such spins so that we can look at large values of the 
    // quantum spin number.

    index = (index)*UnitCell.number + config[i][4] - 1; // uc=1..MAX_UC

    // at this point: 0 <= index < max_number_cells_BT*UnitCell.number
    // note that: 0 <= config[][4] < UnitCell.number

    // sites: (0,0,2,3) -> block1: { [0]=2 [1]=0 [2]=1 [3]=1 }
    //                     block2: { [4]=0 [5]=0 ...         }
    // ToDo: JS replace sites_per_int by block_bits + shift-OPs
    block  = index / sites_per_int;   // 0 .. DESCR_SIZE-2, int position
//  index  = index % sites_per_int;   // 0 .. sites_per_int-1,  bit position
    index  = index - sites_per_int*block; // faster on MIPS?

#ifdef DEBUG  /* makes code bigger, may cause cache misses */
    // better check that in max_index/blocksize in initBTarray (done!)
    if (block >= DESCR_SIZE) {
       // printf("Error: Constant.h DESCR_SIZE <= %d\n", block+1);
       exit(1);
    }
#endif

    // add spin old: + (Params.s2+1)^(site_idx)=(2s+1)^(site_idx)
    //descriptor[block] += divisors[index];
    // add spin old: + 1<<(site_idx*bits_per_site)
    descriptor[block] += 1<<(index*bits_per_site);

#ifdef DEBUG
    if (descriptor[block] < 0) {
        printf("Error descr_to_config: invalid descr[%d]=%d\n"
               " index=%d block_zize=%d\n",
               block, descriptor[block], index, sites_per_int);
    }
    // (0,0,0,1)  -> idx=0     <->    lookUpTable[idx=0..] -> (x,y,z)
    // (0,0,0,uc) -> idx=uc-1 

    static int n_out=0;
    if (Control.output > 6 && n_out<100) { n_out++;
      printf("[%d] (%2d %2d %2d %2d) -> index = %2d -> descr[%d]=%7d\n",
        i, config[i][1],config[i][2],config[i][3],config[i][4],
        index,block,descriptor[block]);
    }
#endif
  }  
  return 0;
}

/*
 *  input:  descriptor[DESCR_SIZE]=list_of_idx_to_xyzc
 *  output: configuration[flips][x,y,z,UCell]   commented by Joerg 2005
 *  called by: obtain, define via findFundamentalSpinConfigurations, ...
 * Oct05: checkBlock removed (obsolete)
 * ToDo: use descriptor[0] to encode size of cluster !?
 */
int findClusterFromDescriptor(int *descriptor, int (*configuration)[5]){
  int i, j, m, index, flips;
  int temp1, temp2, temp;
  static int n_out=0;
  int x, y, z, cell;

  x=y=z=0;   // avoid compiler warnings about using noninitialized vars
  
  if (n_out<=100) n_out++;

  flips=0;  //  the number of spin flips

  for(i=0;i<DESCR_SIZE;i++) {    // DESCR_SIZE*BLOCKLENGTH = SITES*UC
      temp1=descriptor[i];          // block[i] of bits for site.uc
      if (temp1==0) continue;       // no flips coded in that block
#ifdef DEBUG
      if (temp1 <0) {printf("ERROR descr_to_cfg: %d %d\n", i, temp1);break;}
      if (Control.output> 6 && n_out<100)
        printf("descr_to_config descriptor[%d]=%d\n",i,descriptor[i]);
#endif
      for(j=sites_per_int-1;j>=0;j--) {
	
	index=(j)+(i)*sites_per_int;   // decode position: pos*nuc+uc

	// note that: 0 <= index < sites_per_int*DESCR_SIZE

	temp=index;
	index=index/UnitCell.number;       // a/b  site_index  0..maxSite
	cell=temp-UnitCell.number*index;   // a%b  uc_index    0..uc-1

	// 0 <= index < (sites_per_int*DESCR_SIZE)/UnitCell.number 

        // extract bitpattern
	//temp2=temp1 / divisors[j];           // a/b   0..2s numFlips
	//temp1=temp1-temp2*divisors[j];       // a%b   rest
        // 0..2s numFlips on site index (JS0911)
	temp2=(temp1 >> (j*bits_per_site)) & ((1<<bits_per_site)-1);
	
        if (!temp2) continue; // skip if no flips

#ifdef DEBUG
        if (Control.output > 6 && n_out<100)
          printf("i = %2d j = %2d index = %2d -> (%2d %2d %2d %2d) ^ %d ...",
                  i, j, index, lookUpTable[index][0],
                               lookUpTable[index][1],
                               lookUpTable[index][2], cell+1, temp2);
#endif
        x=lookUpTable[index][0];   // see definePointsInBT() input.C
        y=lookUpTable[index][1];   // i=0..sizeLUT-1
        z=lookUpTable[index][2];
        
        for (m=1;m<=temp2;m++) {  // checkBlock: 0 ... Params.s2
          flips++;
          if (flips>MAX_LSUBM) {  // Oct05:  we mean LSUBN here
            printf("Error descr_to_config: flips > MAX_LSUBM\n");
            exit(1);
          }
          configuration[flips][1]=x;
          configuration[flips][2]=y;
          configuration[flips][3]=z;
          configuration[flips][4]=cell+1;
#ifdef DEBUG
          if (Control.output> 4 && n_out<100)
            printf(" site = %2d (%2d %2d %2d %2d)\n",flips,x,y,z,cell+1);
#endif
        }
      }
  }

  return flips;
}

// almost the same as findClusterFromDescriptor (Joerg)
void printOutClusterFromDescriptor(int *descriptor){
  int len;
  static int (*conf)[5]=NULL;
  if (!conf) conf = (int (*)[5]) malloc((MAX_LSUBM+1)*5*sizeof(int));
  len = findClusterFromDescriptor(descriptor, conf);
  if (Params.dim>1) printOutConfig(conf,len); // xy[len]: x y uc, ...\n
  else             printOutPattern(conf,len); // 11..11 or (1.)(.1)
}

// in: descriptor[0..(DESCR_SIZE-1)] = spinconfig of 20 sites
// out: index to spinConfigurations with same descriptor
// Joerg Jun08 ToDo: unificate with findExcitedDescriptorIndex !!!
//     it only differs in [excited]spinConfigurations
// Jun08 JS: numberElements is always 
//           min(NumberConfigurations,MAX_NUM_CONFIGURATIONS)
// JS 0903: unified 3 similar functions
//    findDescriptorIndex()        for spinConfigurations[NumberConfigurations] 
//    findExcitedDescriptorIndex() for excitedSpinConfigurations[NumberExcitedConfigurations] 
//    findDescriptorIndex2()       for latticeAnimals[NumberConfigurations] 
//  numCfgs=1..MAX_NUM_CONFIGURATIONS
int findDescriptorIndex(int *descriptor,
      struct spinConfigs *table, int numCfgs) {

  int i;
  int t1, t2; // upper and lower border for binary search
  int middle;

  if (numCfgs < 1) return 0;

  { // JS0911: could generate a SEGFAULT
    static int err=1;
    if (numCfgs<0 || numCfgs>=MAX_NUM_CONFIGURATIONS) {
      if (err)
        printf("\nERROR findDescriptorIndex: numCfg out of range"
               " (0..MAX_NUM_CONFIGURATIONS != %d)\n",numCfgs);
      err=0; // print once
      numCfgs= MAX_NUM_CONFIGURATIONS; // set to allowed maximum
    }
  }

  // use a binary search and return the `position' of the element (if found!)
  // if NOT found then return zero.

  t1=1-1;
  t2=numCfgs+1;

  // now check that elements in the `middle' do not match `descriptor'
  // Joerg Sep 05: lot of zeros are checked, replace DESCR_SIZE by
  //               a calculated maximum value

  do {
    middle=(t1+t2)/2; // if t1+t2 is odd 
    for(i=DESCR_SIZE-1;i>=0;i--) {
      if (table[middle].descriptor[i]==descriptor[i]) continue;
      if (table[middle].descriptor[i] >descriptor[i]) 
        t2=middle; // new upper bound
      else
        t1=middle; // new lower bound
      break;
    }      
    if (i<0) return middle; 
  } while (t2-t1 > 1);
		       
  // old linear search routine. Useful to keep this as a check of 
  // new binary routine  

  /*
   for(j=numCfgs;j>0;j--) {
     for(i=0;i<DESCR_SIZE;i++)
       if (table[j].descriptor[i]!=descriptor[i]) break;
     if (i==DESCR_SIZE) return j;
   } 
  */
  return 0; // not found
}

#if 1 // ToDo: replaced by split_binary ... soon

/* make binary bitmap spin[]={0..1,0..1,...} from index (Joerg) */
void binary(int index, int length, int *spin)
{
  int i, y, dummy=index;
  if ((int)sizeof(int)*8 < length) { printf("binary Error OVERFLOW"); exit(1); }
  // for(i=1;i<=length;i++,index>>=1) { spin[i] = (index & 1); }
  for(i=1;i<=length;i++) {
    y= dummy/2;
    spin[i]= dummy-2*y;
    dummy=y;
  }
}

/* decode trimary code spin[]={0..2,0..2,...} from index1..3 (Joerg) */
void trimary(int index1, int index2, int index3, int length, int *spin)
{
  int i,y,dummy1, dummy2, dummy3;    
  dummy1=index1; /* spins 21..30 */
  dummy2=index2; /* spins 11..20 */
  dummy3=index3; /* spins  1..10 */
  if (length>30) { printf("trimary Error OVERFLOW"); exit(1); }

  for(i=21;i<=30 && i<=length;i++) {
    y= dummy1/3;
    spin[i]=dummy1-3*y;
    dummy1=y;
  }

  for(i=11;i<=20 && i<=length;i++) {
    y= dummy2/3;
    spin[i]=dummy2-3*y;
    dummy2=y;
  }

  for(i=1;i<=10 && i<=length;i++) {
    y= dummy3/3;
    spin[i]=dummy3-3*y;
    dummy3=y;
  }
}

#endif

